@@ -5,7 +5,7 @@ import pandas as pd
 import os
 from time import sleep
 
-versionNo = "1.0.1"
+versionNo = "1.0.2"
 
 controlEndpoint = "tcp://127.0.0.1:5557"
 dataEndpoint = "tcp://127.0.0.1:5556"
@@ -13,7 +13,7 @@ dataEndpoint = "tcp://127.0.0.1:5556"
 profilesPath = os.path.expanduser('~\\Documents\\Impulse\\Profiles\\')
 profilesPath = profilesPath.replace("\\","/")
 
-
+# Legacy for Impulse version 1.1
 decimalVals = {
     'temperatureMeasured': 6,
     'temperatureSetpoint': 6,
@@ -54,7 +54,7 @@ def createControlMessage(details):
     controlStringB = bytes(controlString, 'utf_8')
     return controlStringB
 
-def sendControl(controlDict):  
+def sendControl(controlDict, showResult=1):  
     controlString = createControlMessage(controlDict)
     context = zmq.Context()
     controlSock = context.socket(zmq.REQ)
@@ -64,18 +64,19 @@ def sendControl(controlDict):
     controlSock.setsockopt(zmq.LINGER, 0)
     controlSock.close()
     context.term()
-    if returnMessage["resultCode"] != "ok":
-        if returnMessage["resultCode"] == "invalidInput":
-            print("[INVALID INPUT]: ", returnMessage["message"], ":")
-            for error in returnMessage["data"]["errors"]:
-                print("    - ", error["parameter"],
-                      error["$type"].split('.')[-1], error["description"])
-        elif returnMessage["resultCode"] == "badRequest":
-            print("[BAD REQUEST]:",
-                  controlString.decode('ascii'))
-        elif returnMessage["resultCode"] == "invalidOperation":
-            print("[INVALID OPERATION]: ",
-                  controlString.decode('ascii'))
+    if showResult==1:
+        if returnMessage["resultCode"] != "ok":
+            if returnMessage["resultCode"] == "invalidInput":
+                print("[INVALID INPUT]: ", returnMessage["message"], ":")
+                for error in returnMessage["data"]["errors"]:
+                    print("    - ", error["parameter"],
+                          error["$type"].split('.')[-1], error["description"])
+            elif returnMessage["resultCode"] == "badRequest":
+                print("[BAD REQUEST]:",
+                      controlString.decode('ascii'))
+            elif returnMessage["resultCode"] == "invalidOperation":
+                print("[INVALID OPERATION]: ",
+                      controlString.decode('ascii'))
     return returnMessage
 
 def disconnect():
@@ -352,9 +353,7 @@ class eventHandler():
         self.eventCatcherThread.start()
 
     def eventCatcher(self):
-        print("Subscribed to impulse events. ")
-        print()
-        sleep(0.01)
+        print("\n Subscribed to impulse events. ")
         while True:
             if self.event.is_set():
                 break
@@ -392,6 +391,32 @@ class eventHandler():
         self.event.set()
 
 
+def getParInfo(channelName, parameterName):
+    controlDict = {
+        '$type' : 'control.impulse.getParameterInformation',
+        'channelName': channelName,
+        'parameterName': parameterName
+        }
+    returnMessage = sendControl(controlDict,0) #Try but do not show invalid operation if the command does not exist
+    if returnMessage['resultCode'] == 'ok':
+        parInfo = returnMessage['data']
+        return parInfo
+    else:
+        return False
+
+def getAllParInfo(channel, header):
+    unitsDict = {}
+    parTypesDict = {}
+    precisionDict = {}
+    for item in header:
+        if item != '$type' and item != 'flags':
+            itemInfo = getParInfo(channel, item)
+            if itemInfo:
+                unitsDict[item]=itemInfo['unit']
+                parTypesDict[item]=itemInfo['type']
+                precisionDict[item]=itemInfo['precision']
+    return [unitsDict, parTypesDict, precisionDict]
+
 class dataHandler():
     def __init__(self, topic):
         self.topic = topic
@@ -404,6 +429,7 @@ class dataHandler():
         self.flags = {}
         self.currentFlag = ""
         self.event = Event()
+        self.precisions = False
 
     def subscribe(self):
         self.context = zmq.Context()
@@ -413,7 +439,7 @@ class dataHandler():
         self.dataSock.setsockopt(zmq.SUBSCRIBE, btopic)
         self.dataThread = Thread(target=self.incomingDataReader)
         self.dataThread.start()
-        print(f"Subscribed to {self.topic} data. ")
+        print(f"\n Subscribed to {self.topic} data. ")
 
     def incomingDataReader(self):
         if self.topic != "massSpec":
@@ -430,6 +456,11 @@ class dataHandler():
                     self.lastMssg = contents
                     if self.header is None:
                         self.header = list(contents.keys())
+                        unitsDict, parTypesDict, precisionDict = getAllParInfo(self.topic, self.header)
+                        if unitsDict:
+                            self.units = unitsDict
+                            self.types = parTypesDict
+                            self.precisions = precisionDict
                         self.data = pd.DataFrame(columns=self.header)
                     self.data = self.data.append(contents, ignore_index=True, sort=False)
         else: # Special DataReceiver for MS data
@@ -442,11 +473,30 @@ class dataHandler():
                 except:
                     continue
                 else:
+                    if self.header is None:
+                        self.units = {}
+                        self.precisions = {}
+                        self.types = {}
+                        for item in contents:
+                            if item != '$type' and item !='channels':
+                                itemInfo = getParInfo('massSpec',item)
+                                if itemInfo:
+                                    self.units[item]=itemInfo['unit']
+                                    self.precisions[item]=itemInfo['precision']
+                                    self.types[item]=itemInfo['type']
+                        channelInfo = getParInfo('massSpec','channels.measuredValue')
+                        if channelInfo:
+                            for channel in contents['channels']:
+                                self.units[channel['name']]=channelInfo['unit']
+                                self.precisions[channel['name']]=channelInfo['precision']
+                                self.types[channel['name']]=channelInfo['type']
+                            
                     for channel in contents["channels"]:
                        contents[channel["name"]]=channel["measuredValue"]
                     del contents["channels"]
                     contents["flags"]=self.currentFlag
                     self.lastMssg = contents
+                    
                     if self.header is None:
                         self.header = list(contents.keys())
                         self.data = pd.DataFrame(columns=self.header)
@@ -464,8 +514,11 @@ class dataHandler():
     def setPointChanged(self):
         self.lastSentSequence = self.lastMssg["sequenceNumber"]
 
-    def roundDataFrame(self, dataSet):       
-        dataSet = dataSet.round(decimalVals)
+    def roundDataFrame(self, dataSet):
+        if self.precisions:
+            dataSet = dataSet.round(self.precisions)
+        else:
+            dataSet = dataSet.round(decimalVals)
         return dataSet
         
     def getLastData(self):
@@ -473,8 +526,13 @@ class dataHandler():
             self.lastSentSequence = self.lastMssg["sequenceNumber"]
             dataSet = self.lastMssg
             for key, value in dataSet.items():
-                if key in decimalVals:
-                    dataSet[key]=round(dataSet[key],decimalVals[key])
+                if self.precisions:
+                    if key in self.precisions:  
+                        if self.types[key]=='Number':
+                            dataSet[key]=round(dataSet[key],self.precisions[key])
+                else:
+                    if key in decimalVals:                    
+                        dataSet[key]=round(dataSet[key],decimalVals[key])   
             return dataSet
 
         else:
@@ -491,8 +549,13 @@ class dataHandler():
             self.lastSentSequence = self.lastMssg["sequenceNumber"]
             dataSet = self.lastMssg
             for key, value in dataSet.items():
-                if key in decimalVals:
-                    dataSet[key]=round(dataSet[key],decimalVals[key])
+                if self.precisions:
+                    if key in self.precisions:
+                        if self.types[key]=='Number':
+                            dataSet[key]=round(dataSet[key],self.precisions[key])
+                else:
+                    if key in decimalVals:                    
+                        dataSet[key]=round(dataSet[key],decimalVals[key])                       
             return dataSet
         else:
             print(f"[ALERT] Please subscribe using impulse.{self.topic}.data.subscribe() first")
@@ -533,5 +596,6 @@ profile = profile()
 heat = heat()
 bias = bias()
 gas = gas()
+stimuli = [heat, bias, gas]
 events = eventHandler()
-events.subscribe()
\ No newline at end of file
+events.subscribe()
